@@ -2,8 +2,8 @@
 # -*- coding: utf-8 -*-
 """
 -------------------------------------------------------------------------------
-input_setup
-Generate input files for a hipims flood model case
+InputHipims
+Generate input files for a hipims flood model
 -------------------------------------------------------------------------------
 @author: Xiaodong Ming
 Created on Tue Mar 31 16:03:57 2020
@@ -23,17 +23,16 @@ __author__ = "Xiaodong Ming"
 import os
 import warnings
 import shutil
-import pickle
-import gzip
+import copy
 import scipy.signal
 import numpy as np
-import matplotlib.pyplot as plt
 from datetime import datetime
 from .Raster import Raster
 from .Boundary import Boundary
 from .ModelSummary import ModelSummary
-from .ModelSummary import _check_rainfall_rate_values
 from .spatial_analysis import sub2map
+from . import indep_functions as indep_f
+from . import rainfall_processing as rp
 #%% grid data for HiPIMS input format
 class InputHipims:
     """To define input files for a HiPIMS flood model case
@@ -44,7 +43,9 @@ class InputHipims:
         num_of_sections: (scalar) number of GPUs to run the model
         Boundary: a boundary object of class Boundary to provide boundary cells
             location, ID, type, and source data.
-        Raster: a raster object to provide DEM data
+        DEM: a Raster object to provide DEM data [alias: Raster].
+        shape: shape of the DEM array
+        header: (dict) header of the DEM grid
         Summary: a ModelSummary object to record model information
         Sections: a list of objects of child-class InputHipimsSub
         attributes_default: (dict) default model attribute names and values
@@ -54,9 +55,12 @@ class InputHipims:
         device_no: (int) the gpu device id(s) to run model
     Properties (Private):
         _valid_cell_subs: (tuple, int numpy) two numpy array indicating rows
-            and cols of valid cells on the DEM grid
+            and cols of valid cells on the DEM array and sorted based on the
+            valid_cell_id, starting from the bottom-left valid cell towards
+            right and up on the DEM array.
         _outline_cell_subs: (tuple, int numpy) two numpy array indicating rows
-            and cols of outline cells on the DEM grid
+            and cols of outline cells on the DEM array and sorted based on the
+            valid_cell_id from small to large
         _global_header: (dict) header of the DEM for the whole model domain
     Methods (public):
         set_parameter: set grid-based parameters
@@ -106,9 +110,12 @@ class InputHipims:
         """
         self.birthday = datetime.now()
         if type(dem_data) is str:
-            self.Raster = Raster(dem_data) # create Raster object
+            self.DEM = Raster(dem_data) # create Raster object
         elif type(dem_data) is Raster:
-            self.Raster = dem_data
+            self.DEM = dem_data
+        self.Raster = self.DEM
+        self.shape = self.DEM.shape
+        self.header = self.DEM.header
         if case_folder is None:
             case_folder = os.getcwd()
         self.case_folder = case_folder
@@ -124,7 +131,7 @@ class InputHipims:
             pass
         else:
             self.__divide_grid()
-            self._global_header = self.Raster.header        
+            self._global_header = self.DEM.header        
         self.set_case_folder() # set data_folders
         self.set_device_no() # set the device number
         self.set_boundary_condition(outline_boundary='fall')
@@ -197,7 +204,7 @@ class InputHipims:
         if parameter_name not in InputHipims.attributes_default.keys():
             raise ValueError('Parameter is not recognized: '+parameter_name)
         if type(parameter_value) is np.ndarray:
-            if parameter_value.shape != self.Raster.array.shape:
+            if parameter_value.shape != self.shape:
                 raise ValueError('The array of the parameter '
                                  'value should have the same '
                                  'shape with the DEM array')
@@ -220,10 +227,10 @@ class InputHipims:
         if rain_mask is None:
             rain_mask = self.attributes['precipitation_mask']
         elif type(rain_mask) is Raster:
-            mask_on_dem = _generate_mask_for_DEM(rain_mask, self.Raster)
+            mask_on_dem = _generate_mask_for_DEM(rain_mask, self.DEM)
             self.attributes['precipitation_mask'] = mask_on_dem.array
         elif type(rain_mask) is np.ndarray:
-            if rain_mask.shape == self.Raster.array.shape:
+            if rain_mask.shape == self.shape:
                 self.attributes['precipitation_mask'] = rain_mask
             else:
                 raise ValueError('The shape of rainfall_mask array '
@@ -237,7 +244,7 @@ class InputHipims:
             warning_str= ('The column of rain source '
                           'is not consistent with the number of rain masks')
             warnings.warn(warning_str)
-        _ = _check_rainfall_rate_values(rain_source, times_in_1st_col=True)
+        _ = rp._check_rainfall_rate_values(rain_source, times_in_1st_col=True)
         self.Summary.add_param_infor('precipitation_mask', rain_mask)
         if rain_source is not None:
             self.attributes['precipitation_source'] = rain_source
@@ -304,8 +311,8 @@ class InputHipims:
             runtime = [0, 3600, 3600, 3600]
         runtime = np.array(runtime)
         self.times = runtime
-        runtime_str = ('start from {0}s, output every {2}s,'+
-                       ' backup every {3}s, and end at {1}s')
+        runtime_str = ('{0}-start, {1}-end, {2}-output interval, '
+                       '{3}-backup interval')
         runtime_str = runtime_str.format(*runtime)
         if hasattr(self, 'Summary'):
             self.Summary.add_items('Run time', runtime_str)
@@ -332,19 +339,21 @@ class InputHipims:
         self.Summary.add_param_infor(param_name, param_value)
 
     def decomposite_domain(self, num_of_sections):
+        obj_mg = copy.deepcopy(self)
         if isinstance(self, InputHipims):
-            self.num_of_sections = num_of_sections
-            self.set_device_no()
-            self._global_header = self.Raster.header
-            self.__divide_grid()
-            self.set_case_folder() # set data_folders
-            outline_boundary = self.Boundary.outline_boundary
-            self.set_boundary_condition(outline_boundary=outline_boundary)
-            self.set_gauges_position()
-            self.Boundary._divide_domain(self)
-            self.birthday = datetime.now()
+            obj_mg.num_of_sections = num_of_sections
+            obj_mg.set_device_no()
+            obj_mg._global_header = obj_mg.Raster.header
+            obj_mg.__divide_grid()
+            obj_mg.set_case_folder() # set data_folders
+            outline_boundary = obj_mg.Boundary.outline_boundary
+            obj_mg.set_boundary_condition(outline_boundary=outline_boundary)
+            obj_mg.set_gauges_position()
+            obj_mg.Boundary._divide_domain(obj_mg)
+            obj_mg.birthday = datetime.now()
         else:
             raise ValueError('The object cannot be decomposited!')
+        return obj_mg
 
     def write_input_files(self, file_tag=None):
         """ Write input files
@@ -358,26 +367,34 @@ class InputHipims:
         """
         self._make_data_dirs()
         grid_files = InputHipims.grid_files
-        if file_tag is None or file_tag == 'all':
-            for grid_file in grid_files: # grid-based files
-                self.write_grid_files(grid_file)
-            self.write_boundary_conditions()
-            self.write_rainfall_source()
-            self.write_gauges_position()
+        if file_tag is None or file_tag == 'all': # write all files
+            file_tag_list = ['z', 'h', 'hU', 'manning', 'sewer_sink',
+                             'cumulative_depth', 'hydraulic_conductivity',
+                             'capillary_head', 'water_content_diff'
+                             'precipitation_mask', 'precipitation_source',
+                             'boundary_condition', 'gauges_pos']
             if self.num_of_sections > 1:
                 self.write_halo_file()
             self.write_mesh_file()
             self.write_runtime_file()
             self.write_device_file()
-        elif file_tag == 'boundary_condition':
-            self.write_boundary_conditions()
-        elif file_tag == 'gauges_pos':
-            self.write_gauges_position()
-        elif file_tag == 'halo':
-            self.write_halo_file()
+        elif type(file_tag) is str:
+            file_tag_list = [file_tag]
+        elif type(file_tag) is list:
+            file_tag_list = file_tag
         else:
-            if file_tag in grid_files:
-                self.write_grid_files(file_tag)
+            print(file_tag_list)
+            raise ValueError(('file_tag should be a string or a list of string'
+                             'from the above list'))
+        for one_file in file_tag_list:
+            if one_file in grid_files: # grid-based files
+                self.write_grid_files(one_file)
+            elif file_tag == 'boundary_condition':
+                self.write_boundary_conditions()
+            elif file_tag == 'precipitation_source':
+                self.write_rainfall_source()
+            elif file_tag == 'gauges_pos':
+                self.write_gauges_position()
             else:
                 raise ValueError('file_tag is not recognized')
 
@@ -422,7 +439,7 @@ class InputHipims:
         rain_source = self.attributes['precipitation_source']
         case_folder = self.case_folder
         num_of_sections = self.num_of_sections
-        write_rain_source(rain_source, case_folder, num_of_sections)
+        indep_f.write_rain_source(rain_source, case_folder, num_of_sections)
         self.Summary.write_readme(self.case_folder+'/readme.txt')
 
     def write_gauges_position(self, gauges_pos=None):
@@ -475,7 +492,7 @@ class InputHipims:
         self._make_data_dirs()
         if is_single_gpu is True or self.num_of_sections == 1:
             file_name = self.data_folders['mesh']+'DEM.txt'
-            self.Raster.write_asc(file_name)
+            self.DEM.write_asc(file_name)
         else:
             for obj_section in self.Sections:
                 file_name = obj_section.data_folders['mesh']+'DEM.txt'
@@ -488,35 +505,35 @@ class InputHipims:
         """
         if time_values is None:
             time_values = self.times
-        write_times_setup(self.case_folder, self.num_of_sections, time_values)
+        indep_f.write_times_setup(self.case_folder, self.num_of_sections,
+                                  time_values)
     
     def write_device_file(self, device_no=None):
         """Create device_setup.dat for choosing GPU number to run the model
         """
         if device_no is None:
             device_no = self.device_no
-        write_device_setup(self.case_folder, self.num_of_sections, device_no)
+        indep_f.write_device_setup(self.case_folder, self.num_of_sections,
+                                   device_no)
 
     def save_object(self, file_name):
         """ Save object as a pickle file
         """
-        if not file_name.endswith('.pickle'):
-            file_name = file_name+'.pickle'
-        save_object(self, file_name, compression=True)
+        indep_f.save_object(self, file_name, compression=True)
 #------------------------------------------------------------------------------
 #******************************* Visualization ********************************
 #------------------------------------------------------------------------------
     def domain_show(self, figname=None, dpi=None, **kwargs):
         """Show domain map of the object
         """
-        fig, ax = self.Raster.mapshow(**kwargs)
+        fig, ax = self.DEM.mapshow(**kwargs)
         cell_subs = self.Boundary.cell_subs
         legends = []
         num = 0
         for cell_sub in cell_subs:
             rows = cell_sub[0]
             cols = cell_sub[1]
-            X, Y = sub2map(rows, cols, self.Raster.header)
+            X, Y = sub2map(rows, cols, self.DEM.header)
             ax.plot(X, Y, '.')
             legends.append('Boundary '+str(num))
             num = num+1
@@ -526,47 +543,35 @@ class InputHipims:
         if figname is not None:
             fig.savefig(figname, dpi=dpi)
         return fig, ax
+    
+    def plot_rainfall_map(self, cellsize=None, method='sum', **kw):
+        """
+        """
+        rain_source = self.attributes['precipitation_source']
+        rain_mask = self.attributes['precipitation_mask']
+        if rain_mask.size == 1:
+            rain_mask = self.DEM.array*0+rain_mask
+        rain_mask_obj = Raster(array=rain_mask, header=self.header)
+        rain_map_obj = rp.get_spatial_map(rain_source, rain_mask_obj, 
+                                          cellsize, method)
+        rain_map_obj.mapshow(**kw)
 
-    def plot_rainfall_source(self,start_date=None, method='mean'):
+    def plot_rainfall_curve(self, start_date=None, method='mean'):
         """ Plot time series of average rainfall rate inside the model domain
         start_date: a datetime object to give the initial date and time of rain
         method: 'mean'|'max','min','mean'method to calculate gridded rainfall 
         over the model domain
-            
-        """        
+        """
         rain_source = self.attributes['precipitation_source']
         rain_mask = self.attributes['precipitation_mask']
-        if type(rain_mask) is np.ndarray:
-            rain_mask = rain_mask[~np.isnan(self.Raster.array)]
-        rain_mask = rain_mask.astype('int32')
-        rain_mask_unique = np.unique(rain_mask).flatten()
-        rain_mask_unique = rain_mask_unique.astype('int32')
-        
-        rain_source_valid = rain_source[:,rain_mask_unique+1]
-        time_series = rain_source[:,0]
-        if type(start_date) is datetime:
-            from datetime import timedelta
-            time_delta = np.array([timedelta(seconds=i) for i in time_series])
-            time_x = start_date+time_delta
+        rain_mask = np.array(rain_mask)
+        if rain_mask.size == 1:
+            rain_mask = self.DEM.array*0+rain_mask
         else:
-            time_x = time_series
-        if method == 'mean':
-            value_y = np.mean(rain_source_valid,axis=1)
-        elif method== 'max':
-            value_y = np.max(rain_source_valid,axis=1)
-        elif method== 'min':
-            value_y = np.min(rain_source_valid,axis=1)
-        elif method== 'median':
-            value_y = np.median(rain_source_valid,axis=1)
-        else:
-            raise ValueError('Cannot recognise the calculation method')
-        value_y =  value_y*3600*1000
-        plot_data = np.c_[time_x,value_y]
-        fig, ax = plt.subplots()
-        ax.plot(time_x,value_y)
-        ax.set_ylabel('Rainfall rate (mm/h)')
-        ax.grid(True)
-        plt.show()
+            rain_mask[np.isnan(self.DEM.array)] = np.nan
+        plot_data = rp.get_time_series(rain_source, rain_mask, start_date, 
+                                    method=method)
+        rp.plot_time_series(plot_data, method=method)
         return plot_data
 #------------------------------------------------------------------------------
 #*************************** Protected methods ********************************
@@ -579,7 +584,7 @@ class InputHipims:
         _outline_cell_subs
         """
         if dem_array is None:
-            dem_array = self.Raster.array
+            dem_array = self.DEM.array
         valid_id, outline_id = _get_cell_id_array(dem_array)
         subs = np.where(~np.isnan(valid_id))
         id_vector = valid_id[subs]
@@ -604,12 +609,12 @@ class InputHipims:
             if self.num_of_sections == 1:
                 return 1 # do not divide if num_of_sections is 1
         num_of_sections = self.num_of_sections
-        dem_header = self.Raster.header
+        dem_header = self.header
         self._global_header = dem_header
         # subscripts of the split row [0, 1,...] from bottom to top
-        split_rows = _get_split_rows(self.Raster.array, num_of_sections)
+        split_rows = _get_split_rows(self.DEM.array, num_of_sections)
         array_local, header_local = \
-            _split_array_by_rows(self.Raster.array, dem_header, split_rows)
+            _split_array_by_rows(self.DEM.array, dem_header, split_rows)
         # to receive InputHipimsSub objects for sections
         Sections = []
         section_sequence = np.arange(num_of_sections)
@@ -660,7 +665,7 @@ class InputHipims:
                             or a list of vectors for each sub domain
         """
         # get grid value
-        dem_shape = self.Raster.array.shape
+        dem_shape = self.shape
         grid_values = np.zeros(dem_shape)
         if add_initial_water:
             add_value = 0.0001
@@ -682,7 +687,7 @@ class InputHipims:
             return grid_values
         # set grid value for the entire domain
         if attribute_name == 'z':
-            grid_values = self.Raster.array
+            grid_values = self.DEM.array
         elif attribute_name == 'h':
             grid_values = grid_values+self.attributes['h0']
             # traversal each boundary to add initial water
@@ -844,7 +849,7 @@ class InputHipims:
                                                     cell_subs[1], 1)
                         theta = np.arctan(boundary_slope[0])
                         boundary_length = cell_subs[0].size* \
-                                          self.Raster.header['cellsize']
+                                          self.DEM.header['cellsize']
                         hUx = hU_source[:, 1]*np.cos(theta)/boundary_length
                         hUy = hU_source[:, 1]*np.sin(theta)/boundary_length
                         hU_source = np.c_[hU_source[:, 0], hUx, hUy]
@@ -858,7 +863,6 @@ class InputHipims:
 
     def __write_gauge_pos(self, file_folder):
         """write monitoring gauges
-        Private version of write_gauge_position
         gauges_pos.dat
         file_folder: folder to write file
         gauges_pos: 2-col numpy array of X and Y coordinates
@@ -874,7 +878,6 @@ class InputHipims:
 
     def __write_gauge_ind(self, file_folder):
         """write monitoring gauges index for mult-GPU sections
-        Private function of write_gauge_position
         gauges_ind.dat
         file_folder: folder to write file
         gauges_ind: 1-col numpy array of index values
@@ -1114,15 +1117,6 @@ def _create_io_folders(case_folder, make_dir=False):
                     'mesh':dir_mesh, 'field':dir_field}
     return data_folders
 
-def _check_case_folder(case_folder):
-    """ check the format of case folder
-    """
-    if case_folder is None:
-        case_folder = os.getcwd()
-    if not case_folder.endswith('/'):
-        case_folder = case_folder+'/'
-    return case_folder
-
 def _generate_mask_for_DEM(mask_origin, dem_data):
     """Interpolate orginal mask file to the dem data
     the interpolating method is nearest
@@ -1143,102 +1137,43 @@ def _generate_mask_for_DEM(mask_origin, dem_data):
                          'string or a Raster object')
     mask_on_dem = dem_obj.grid_interpolate(mask_obj)
     return mask_on_dem
-        
-#%% ***************************************************************************
-# *************************Public functions************************************
-def write_times_setup(case_folder=None, num_of_sections=1, time_values=None):
-    """
-    Generate a times_setup.dat file. The file contains numbers representing
-    the start time, end time, output interval, and backup interval in seconds
-    time_values: array or list of int/float, representing time in seconds,
-        default values are [0, 3600, 1800, 3600]
-    """
-    case_folder = _check_case_folder(case_folder)
-    if time_values is None:
-        time_values = np.array([0, 3600, 1800, 3600])
-    time_values = np.array(time_values)
-    time_values = time_values.reshape((1, time_values.size))
-    if num_of_sections == 1:
-        np.savetxt(case_folder+'/input/times_setup.dat', time_values, fmt='%g')
-    else:
-        np.savetxt(case_folder+'/times_setup.dat', time_values, fmt='%g')
-    print('times_setup.dat created')
 
-def write_device_setup(case_folder=None,
-                       num_of_sections=1, device_values=None):
+def copy_input_obj(obj_in):
+    """Copy the an object of class InputHipims
     """
-    Generate a device_setup.dat file. The file contains numbers representing
-    the GPU number for each section
-    case_folder: string, the path of model
-    num_of_sections: int, the number of GPUs to use
-    device_values: array or list of int, representing the GPU number
-    """
-    case_folder = _check_case_folder(case_folder)
-    if device_values is None:
-        device_values = np.array(range(num_of_sections))
-    device_values = np.array(device_values)
-    device_values = device_values.reshape((1, device_values.size))
-    if num_of_sections == 1:
-        np.savetxt(case_folder+'/input/device_setup.dat',
-                   device_values, fmt='%g')
-    else:
-        np.savetxt(case_folder+'/device_setup.dat', device_values, fmt='%g')
-    print('device_setup.dat created')
-
-def write_rain_source(rain_source, case_folder=None, num_of_sections=1):
-    """ Write rainfall sources [Independent function from hipims class]
-    rain_source: numpy array, The 1st column is time in seconds, the 2nd
-        towards the end columns are rainfall rate in m/s for each source ID in
-        rainfall mask array
-    if for multiple GPU, then copy the rain source file to all domain folders
-    case_folder: string, the path of model
-    """
-    rain_source = np.array(rain_source)
-    # check rainfall source value to avoid very large raifall rates
-    _ = _check_rainfall_rate_values(rain_source)
-    case_folder = _check_case_folder(case_folder)
-    fmt1 = '%g'  # for the first col: times in seconds
-    fmt2 = '%.8e'  # for the rest array for rainfall rate m/s
-    num_mask_cells = rain_source.shape[1]-1
-    format_spec = [fmt2]*num_mask_cells
-    format_spec.insert(0, fmt1)
-    if num_of_sections == 1: # single GPU
-        file_name = case_folder+'input/field/precipitation_source_all.dat'
-    else: # multi-GPU
-        file_name = case_folder+'0/input/field/precipitation_source_all.dat'
-    with open(file_name, 'w') as file2write:
-        file2write.write("%d\n" % num_mask_cells)
-        np.savetxt(file2write, rain_source, fmt=format_spec, delimiter=' ')
+    
+    dem_data = Raster(array=obj_in.Raster.array, 
+                             header=obj_in.Raster.header)
+    num_of_sections = obj_in.num_of_sections
+    case_folder = obj_in.case_folder
+    # initialize a new object
+    obj_copy = InputHipims(dem_data=dem_data, num_of_sections=num_of_sections,
+                           case_folder=case_folder)
+    attr_dict = copy.deepcopy(obj_in.__dict__)
+    if 'Raster' in attr_dict.keys():
+        del attr_dict['Raster']
+    del attr_dict['Boundary']
+    del attr_dict['Summary']
     if num_of_sections > 1:
-        for i in np.arange(1, num_of_sections):
-            field_dir = case_folder+str(i)+'/input/field/'
-            shutil.copy2(file_name, field_dir)
-    print('precipitation_source_all.dat created')
-
-def load_object(file_name):
-    """ Read a pickle file as an InputHipims object
-    """
-    #read an InputHipims object file
-    try:
-        with gzip.open(file_name, 'rb') as input_file:
-            obj = pickle.load(input_file)
-    except:
-        with open(file_name, 'rb') as input_file:
-            obj = pickle.load(input_file)
-    print(file_name+' loaded')
-    return obj
-
-def save_object(obj, file_name, compression=True):
-    """ Save the object
-    """
-    # Overwrites any existing file.
-    if compression:
-        with gzip.open(file_name, 'wb') as output_file:
-            pickle.dump(obj, output_file, pickle.HIGHEST_PROTOCOL)
-    else:
-        with open(file_name, 'wb') as output_file:
-            pickle.dump(obj, output_file, pickle.HIGHEST_PROTOCOL)
-    print(file_name+' has been saved')
+        del attr_dict['Sections']
+    # set object attributes
+    for key in attr_dict.keys():
+        obj_copy.__dict__[key] = copy.deepcopy(attr_dict[key])
+    # set boundary conditions
+    boundary_list = obj_in.Boundary.boundary_list
+    outline_boundary = obj_in.Boundary.outline_boundary
+    obj_copy.set_boundary_condition(boundary_list, outline_boundary)
+    # copy summary
+    summary_infor = obj_in.Summary.information_dict
+    infor_keys = list(summary_infor.keys())
+    iloc = 0
+    for i in range(len(infor_keys)):
+        if infor_keys[i] == 'gauges_pos':
+            iloc = i
+    infor_keys = infor_keys[iloc+1:]
+    for key in infor_keys:
+        obj_copy.Summary.add_items(key, summary_infor[key])
+    return obj_copy
 
 def main():
     print('Class to setup input data')
