@@ -11,12 +11,11 @@ Created on Wed Apr  1 23:46:54 2020
 import os
 import datetime
 import warnings
-import pickle
-import gzip
 import numpy as np
 import pandas as pd
 import hipims_io.spatial_analysis as sp
 from .Raster import Raster
+from .indep_functions import save_object
 class OutputHipims:
     """To read and analyze otuput files from a HiPIMS flood model
     Properties (public):
@@ -285,33 +284,6 @@ class OutputHipims:
         """
         save_object(self, file_name, compression=True)
     
-def load_object(file_name):
-    """ Read a pickle file as an InputHipims object
-    """
-    #read an object file
-    try:
-        with gzip.open(file_name, 'rb') as input_file:
-            obj = pickle.load(input_file)
-    except:
-        with open(file_name, 'rb') as input_file:
-            obj = pickle.load(input_file)
-    print(file_name+' loaded')
-    return obj
-
-def save_object(obj, file_name, compression=True):
-    """ Save the object
-    """
-    # Overwrites any existing file.
-    if not file_name.endswith('.pickle'):
-        file_name = file_name+'.pickle'
-    if compression:
-        with gzip.open(file_name, 'wb') as output_file:
-            pickle.dump(obj, output_file, pickle.HIGHEST_PROTOCOL)
-    else:
-        with open(file_name, 'wb') as output_file:
-            pickle.dump(obj, output_file, pickle.HIGHEST_PROTOCOL)
-    print(file_name+' has been saved')
-    
 #%% =======================Supporting functions===============================
 def _combine_gauges_data_via_ind(case_folder, num_section, file_tag):
     """Combine gauges outputs from multi-gpu models according to gauges
